"""Utils, like logger
"""
from loguru import logger
from loguru._logger import Logger
from random_word import RandomWords


global log_me
log_me: Logger = logger
log_me.disable('bgameb')

rw = RandomWords()


def log_enable(
    log_path: str = './logs/game.log',
    log_level: str = 'DEBUG'
        ) -> None:
    """Enable logging

    Args:
        log_path (str, optional): path to log file.
                                  Defaults to './logs/game.log'.
        log_level (str, optional): logging level. Defaults to 'DEBUG'.
    """
    log_me.remove()
    log_me.add(
        sink=log_path,
        level=log_level,
        format='{extra[classname]}: "{extra[name]}" -> func {function} | ' +
        '{time:YYYY-MM-DD at HH:mm:ss} | {level} | {message}',
    )
    log_me.enable('bgameb')


def get_random_name(max_lenght: int = 10) -> str:
    """Get english word of given len

    Returns:
        str: random word
    """
    def roll():

        result = rw.get_random_word(
            minLength=6,
            maxLength=max_lenght,
            includePartOfSpeech='noun',
            )
        if not result:
            result = roll()

        return result

    return roll()



"""Main engine to create games
"""
from typing import Optional, Literal
from dataclasses import dataclass, field
from bgameb.tools import TOOLS, TOOLS_TYPES
from bgameb.stuff import STUFF, STUFF_TYPES
from bgameb.errors import ComponentClassError
from bgameb.constructs import Components, BaseGame


@dataclass
class Game(BaseGame):
    """Create the game object
    """
    name: Optional[str] = None
    stuff: Components = field(default_factory=Components, init=False)
    tools: Components = field(default_factory=Components, init=False)

    def __post_init__(self) -> None:
        self.stuff = Components()
        self.tools = Components()
        super().__post_init__()

    def add(
        self,
        component: Literal[STUFF_TYPES, TOOLS_TYPES],
        name: Optional[str] = None,
            ) -> None:
        if component in STUFF.keys():
            self.stuff.add(STUFF[component], name=name)
            self.logger.info(
                f'{component} is added: {self.stuff.get_names()}.'
                )
        elif component in TOOLS.keys():
            self.tools.add(TOOLS[component], _game=self, name=name)
            self.logger.info(
                f'{component} is added: {self.tools.get_names()}.'
                )
        else:
            raise ComponentClassError(component, self.logger)



"""Game tools classes like shakers or decks
"""
from typing import Optional, Tuple, Dict, Literal
from dataclasses import dataclass, field
from bgameb.stuff import Roller, Card
from bgameb.constructs import BaseTool


@dataclass
class Shaker(BaseTool):
    """Create shaker for roll dices or flip coins
    """
    name: Optional[str] = None
    last: Dict[str, Tuple[int]] = field(default_factory=dict, init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.last = {}
        self._stuff_to_add = Roller

    def roll(self) -> Dict[str, Tuple[int]]:
        """Roll all stuff with shaker and return results

        Return:
            Dict[str, Tuple[int]]: result of roll

        .. code-block::
            :caption: Example:

                {
                    "six_dice": (5, 3, 2, 5),
                    "twenty_dice": {2, 12, 4},
                }
        """
        roll = {}

        for name, roller in self.stuff.items():
            roll[name] = roller.roll()

        self.last = roll
        self.logger.debug(f'Result of roll: {roll}')

        return self.last


@dataclass
class Deck(BaseTool):
    """Create deck for cards
    """
    name: Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        self._stuff_to_add = Card

    def deal(self) -> None:
        """Deal deck for play from deck_cards
        """
        raise NotImplementedError

    def shuffle(self) -> None:
        """Shuffle in/out deal_cards
        """
        raise NotImplementedError

    def arrange(self) -> None:
        """Arrange in/out deal_cards
        """

    def look(self) -> None:
        """Look cards in in/out deal_cards
        """
        raise NotImplementedError

    def pop(self) -> None:
        """Pop cards from in to out or visa versa
        """
        raise NotImplementedError

    def search(self, name: str) -> None:
        """Search for cards in in/out deal_cards

        Args:
            name (str): name of card
        """
        raise NotImplementedError

    def move(self) -> None:
        """Move stuffs from this deck toi another
        """
        raise NotImplementedError


TOOLS = {
    'shaker': Shaker,
    'deck': Deck,
}
TOOLS_TYPES = Literal['roller', 'card']



"""Game dices, coins, cards and other stuffs
"""
import random
from typing import List, Optional, Literal
from dataclasses import dataclass, field
from dataclasses_json import config
from bgameb.constructs import BaseStuff
from bgameb.errors import StuffDefineError


@dataclass
class RollerType(BaseStuff):
    """Base class for define types of rollers or fliped objects

    If name isn't given - it is randomly generated.

    Define name to identify later this object by unique name.
    For example: 'six_side_dice'

    Sides attr define number of sides of roller. Default to 2.
    Sides can't be less than 2, because one-sided roller is
    strongly determined and 0zero-sided is imposible.

    .. code-block::
        :caption: Example:

            dice = RollerType(name='coin', sides=2)

    Raises:
        StuffDefineError: number of sides less than 2
    """
    name: Optional[str] = None
    sides: int = 2

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.sides < 2:
            raise StuffDefineError(
                message=f'Number of sides={self.sides} '
                        f'for {self.name}. Needed >= 2.',
                logger=self.logger
                )


@dataclass
class Roller(RollerType):
    """Creare the roller (like dice or coin)
    """
    _range: List[int] = field(
        default_factory=list,
        init=False,
        metadata=config(exclude=lambda x: True),
        repr=False
        )
    count: int = 0

    def __post_init__(self) -> None:
        super().__post_init__()
        self._range = list(range(1, self.sides + 1))

    def roll(self) -> List[int]:
        """Roll and return result

        Returns:
            List[int]: result of roll
        """
        roll = [random.choices(self._range, k=1)[0] for _ in list(range(self.count))]
        self.logger.debug(f'Is rolled {roll=}')
        return roll


@dataclass
class CardType(BaseStuff):
    """Create the card

    If name isn't given - it is randomly generated.

    Define name to identify later this object by unique name.
    For example: 'unique_card'

    .. code-block::
        :caption: Example:

            card = CardType(name='unique_card')
    """
    name: Optional[str] = None
    open: bool = False
    tapped: bool = False
    side: Optional[str] = None

    def __post_init__(self) -> None:
        super().__post_init__()


@dataclass
class Card(CardType):
    """Create the card
    """

    def __post_init__(self) -> None:
        super().__post_init__()

    def flip(self) -> None:
        """Face up or face down the card regardles of it condition
        """
        if self.open:
            self.open = False
            self.logger.debug(f'Card face down.')
        else:
            self.open = True
            self.logger.debug(f'Card face up.')

    def face_up(self) -> None:
        """Face up the card and return text
        """
        self.open = True
        self.logger.debug(f'Card face up.')

    def face_down(self) -> None:
        """Face down the card
        """
        self.open = False
        self.logger.debug(f'Card face down.')

    def tap(self, side='right') -> None:
        """Tap the card to the given side
        """
        self.tapped = True
        self.side = side
        self.logger.debug(f'Card taped to side {side}.')

    def untap(self) -> None:
        """Untap the card
        """
        self.tapped = False
        self.logger.debug(f'Card untaped.')

    def alter(self) -> None:
        """Many cards have alter views. For example
        card can have main view, that apply most time of the game
        and second view, that apply only if card played as
        that alternative. For ease of understanding, consider that
        different views of the same card are not related directly
        to each other.
        """
        raise NotImplementedError

    def attach(self) -> None:
        """Some rules can attach any stuff to card
        """
        raise NotImplementedError


STUFF = {
    'roller': RollerType,
    'card': CardType,
}
STUFF_TYPES = Literal['roller', 'card']



"""Game players classes
"""
from dataclasses import dataclass
from bgameb.constructs import BasePlayer


@dataclass
class Human(BasePlayer):
    """Base class to create a human palyer
    """


@dataclass
class Bot(BasePlayer):
    """Base class to create a bot player
    """



"""Custom error classes
"""

class CustomRuntimeError(RuntimeError):
    """Base class for other runtime exceptions
    """


class ComponentNameError(CustomRuntimeError):
    """Component with given name is in Components
    class instance.
    """
    def __init__(self, name) -> None:
        self.message = f'Component with {name=} is exist in ' + \
                        'Components class instance'
        super().__init__(self.message)


class ComponentClassError(CustomRuntimeError):
    """Given class isnt component
    """
    def __init__(self, obj_, logger) -> None:
        self.message = f'Given: {obj_} not a component.'
        logger.exception(self.message)
        super().__init__(self.message)


class StuffDefineError(AttributeError):
    """Defining stuff error
    """
    def __init__(self, message, logger) -> None:
        self.message = message
        logger.exception(self.message)
        super().__init__(self.message)



"""Base constructs for build package objects
"""
from typing import (
    Dict, List, Optional, Any, Iterable, Sequence
    )
from abc import ABC, abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass, field
from dataclasses_json import DataClassJsonMixin, config
from bgameb.errors import ComponentNameError, StuffDefineError
from bgameb.utils import log_me, get_random_name


class Components(Mapping):
    """Components mapping.

    This class inherit from dict and not implemets
    __setattr__ and __setitems__. Class represents
    additional dot-acces for attributes of dict
    """
    def __init__(
        self,
        *args,
        **kwargs
            ) -> None:
        """Args must be a dicts
        """
        super().__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                self.__dict__.update(arg)
            else:
                raise AttributeError('Args must be a dicts')
        if kwargs:
            self.__dict__.update(kwargs)

    def __iter__(self) -> Iterable:
        return iter(self.__dict__)

    def __getattr__(self, attr: str):
        try:
            return self.__dict__[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr: str, value) -> None:
        raise NotImplementedError('This method not implementd for Components')

    def __delattr__(self, attr: str) -> None:
        try:
            del self.__dict__[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setitem__(self, key: str, value) -> None:
        raise NotImplementedError('This method not implementd for Components')

    def __getitem__(self, key: str):
        return self.__dict__[key]

    def __delitem__(self, key: str) -> None:
        del self.__dict__[key]

    def __repr__(self):
        items = (f"{k}={v!r}" for k, v in self.__dict__.items())
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __len__(self) -> int:
        return len(self.__dict__)

    def _chek_in(self, name: str) -> Optional[bool]:
        """Chek is name of component is unique

        Args:
            name (str): name for component

        Raises:
            ComponentNameError: name id not unique

        Returns:
            Optional[bool]: if it is in
        """
        if name in self.__dict__.keys():
            raise ComponentNameError(name=name)
        return True

    def _update(
        self,
        component,
        kwargs: Dict[str, Any]
            ) -> None:
        """Update components dict

        Args:
            component: component class
            kwargs (Dict[str, Any]): aditional args
        """
        self.__dict__.update(
            {kwargs['name']: component(**kwargs)}
            )

    def add(self, component, **kwargs) -> None:
        """Add component to Components dict. Components with
        same names as existed cant be added.

        Raises:
            ComponentNameError: name id not unique

        Args:
            component (Component): component class
            kwargs: aditional args
        """
        if kwargs.get('name'):
            self._chek_in(kwargs['name'])
        else:
            self._chek_in(component.name)
            kwargs['name'] = component.name
        self._update(component, kwargs)

    def add_replace(self, component, **kwargs) -> None:
        """Add or replace component in Components dict.

        Args:
            component (Component): component class
            kwargs: aditional args
        """
        if not kwargs.get('name'):
            kwargs['name'] = component.name
        self._update(component, kwargs)

    def get_names(self) -> List[str]:
        """Get names of all components of class

        Returns:
            List[str]: lisct of names of conatined components
        """
        return list(self.__dict__)


class CardTexts(dict):
    """Cards texts collection
    """
    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)

    def __getattr__(self, attr: str) -> str:
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(attr)

    def __setattr__(self, attr: str, value: str) -> None:
        self[attr] = value

    def __delattr__(self, attr: str) -> None:
        del self[attr]

    def __repr__(self):
        items = (f"{k}={v!r}" for k, v in self.items())
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __eq__(self, other: Any) -> bool:
        if isinstance(self, dict) and isinstance(other, dict):
            return self.__dict__ == other.__dict__
        return NotImplemented


@dataclass
class Base(DataClassJsonMixin, ABC):
    """Base class for game, stuff and tools

    Inherited classes needs attr name implementation
    """

    def __post_init__(self) -> None:
        # set random name
        if not self.name:
            self.name = get_random_name()

        # set logger
        self.logger = log_me.bind(
            classname=self.__class__.__name__,
            name=self.name)
        self.logger.info(
            f'{self.__class__.__name__} created with {self.name=}.'
            )

    @staticmethod
    def _chek_name(name: str, chek: Sequence[str]) -> bool:
        """Chek exist name string in seequence
        # TODO: test me

        Args:
            name (str): name sting
            check (Sequence[str]): cheked collection

        Returns:
            bool: name is exist or not in collection
        """
        if name in chek:
            return True
        return False


@dataclass
class BaseGame(Base, ABC):
    """Base class for game

    Inherited classes needs attr name implementation
    """

    @abstractmethod
    def add(self, component: str, name: Optional[str] = None) -> None:
        """Add stuff or tools to game

        Args:
            component (str): stuff or tool type
            name (str, optional): name of added component.
                                  Defaults to None.
        """


@dataclass
class BaseStuff(Base, ABC):
    """Base class for game stuff (like dices or cards)

    Inherited classes needs attr name implementation
    """


@dataclass
class BaseTool(Base, ABC):
    """Base class for game tools (like decks or shakers)

    Inherited classes needs attr name implementation
    """
    _game: BaseGame = field(
        metadata=config(exclude=lambda x: True),
        repr=False
        )
    _stuff_to_add: BaseStuff = field(
        metadata=config(exclude=lambda x: True),
        repr=False,
        init=False
        )
    stuff: Components = field(default_factory=Components, init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.stuff = Components()

    def add(self, name: str, count: int = 1) -> None:
        """Add stuff to the tool stuff collection

        Args:
            name (str): name of stuff
            count (int, optional): count of stuff copy Defaults to 1.

        Raises:
            StuffDefineError: count of stuff nonpositive
                              or stuff not exist
        """
        if not self._chek_name(name, self._game.stuff.keys()):
            raise StuffDefineError(
                message=f"Stuff with {name=} not exist in a game.",
                logger=self.logger
                )

        if count < 1:
            raise StuffDefineError(
                message=f"Can't add {count} stuff.",
                logger=self.logger
                )

        # add roller and set a count
        if name not in self.stuff.get_names():
            self.stuff.add(self._stuff_to_add, **self._game.stuff[name].to_dict())
            self.stuff[name].count = count
            self.logger.debug(
                f'Added stuff with "{name=}" and {count=}.'
                )

        # if exist increase a count
        else:
            self.stuff[name].count += count
            self.logger.debug(
                f'Number of stuff with "{name=}" increased by {count}. ' +
                f'Result count is {self.stuff[name].count}'
                )

    def _decrease(self, name: str, count: int) -> None:
        """Decrease number of stuff by count

        Args:
            name (str): name of stuff
            count (int): count
        """
        self.stuff[name].count -= count
        self.logger.debug(
            f'Removed {count} stuff with {name=}. ' +
            f'Result count is {self.stuff.get(name, 0)}'
            )
        if self.stuff[name].count <= 0:
            del self.stuff[name]

    def remove(
        self,
        name: Optional[str] = None,
        count: Optional[int] = None
            ) -> None:
        """Remove any kind of stuff copy from tool by its name and count

        Args:
            name (str, optional): name of stuff. Defaults to None.
            count (int, optional): count of stuff. Defaults to None.

        You can use any combination for names and counts to remove all
        stuff, all or any by name or by count.


        Raises:
            StuffDefineError: nonpositive integer for count
        """
        if count is not None and count <= 0:
            raise StuffDefineError(
                f"Count must be a integer greater than 0.",
                logger=self.logger
                )

        if name is None:

            if count is None:
                self.stuff = Components()
                self.logger.debug('Removed all stuff')
            else:
                for stuff in list(self.stuff.keys()):
                    self._decrease(stuff, count)

        elif self.stuff.get(name):
            if count is None:
                del self.stuff[name]
                self.logger.debug(
                    f'Removed all stuff with {name=}.'
                    )
            else:
                self._decrease(name, count)

        else:
            raise StuffDefineError(
                message=f"Stuff with {name=} not exist in tool.",
                logger=self.logger
                )


@dataclass
class BasePlayer(Base):
    """Base class for game players and bots

    Inherited classes needs attr name implementation
    """



from typing import List
import os, pathlib
from setuptools import setup, find_packages


def strip_comments(string: str) -> str:
    """Remove comments string

    Args:
        string (str): some string

    Returns:
        str: striped string
    """
    return string.split('#', 1)[0].strip()

def get_dependencies(*req: str) -> List[str]:
    """Get dependencies from requirements.txt

    Returns:
        List[str]: list of dependencies strings
    """
    return list(filter(
        None,
        [strip_comments(l) for l in open(
            os.path.join(os.getcwd(), *req)
        ).readlines()]
    ))

here = pathlib.Path(__file__).parent.resolve()
NAME = 'bgameb'
AUTHOR = 'Konstantin Klepikov'
EMAIL = 'oformleno@gmail.com'
DESCRIPTION = 'Board Game Builder'
LONG_DESCRIPTION = (here / "README.md").read_text(encoding="utf-8")
SOURCE_URL = 'https://github.com/KonstantinKlepikov/BoardGameBuilder'
DOCS = 'https://konstantinklepikov.github.io/BoardGameBuilder/'

setup(
    name=NAME,
    install_requires=get_dependencies('requirements.txt'),
    extras_require={
        "dev": get_dependencies('requirements-dev.txt'),
    },
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    url=SOURCE_URL,
    project_urls={
        'Docs': DOCS,
        'Source': SOURCE_URL,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="framework",
    license='MIT',
    python_requires='>=3.8.10',
    packages=find_packages(exclude=('tests*',)),
    use_incremental=True,
)